﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Diagnostics;

//using Microsoft.Win32;

using MTCMS;

namespace MTCMSNETDemo
{
    public partial class MainWindow: Window
    {
        protected const int BIG_BLOCK_DATA_SIZE = 900;
        protected const int BIG_BLOCK_DATA_SMALL_SIZE = 45;

        // Selection Type
        private static byte SELECTION_TYPE_APPLICATION = 0x00;
        private static byte SELECTION_TYPE_LANGUAGE = 0x01;

        // Selection Status
        private static byte SELECTION_STATUS_COMPLETED = 0x00;
        private static byte SELECTION_STATUS_CANCELLED = 0x01;
        private static byte SELECTION_STATUS_TIMED_OUT = 0x02;

        private int m_emvARCType;

        private bool m_enableSwipe = false;
        private bool m_enableChip = true;
        private bool m_enableContactless = false;

        private byte[] ApprovedARC = new byte[] { ( byte ) 0x8A, 0x02, 0x30, 0x30 };
        private byte[] DeclinedARC = { ( byte ) 0x8A, 0x02, 0x30, 0x35 };

        private int mSubscriptionOption = 0;

        private int mMSRRequestOption = 0;

        private MTDevice mDevice;

        private MTConnectionType mConnectionType;

        private MTConnectionState mConnectionState;

        protected string mPINEntryDisplay = "";

        protected string mBaudRate = "9600";
        protected string mDataBits = "8";
        protected string mParity = "NONE";
        protected string mStopBits = "1";
        protected string mHandshake = "NONE";
        protected string mStartingByte = "";
        protected string mEndingByte = "0x0A";
        protected bool mCRC = false;

        protected int mReceivingBigBlockDataTotalLength;
        protected int mReceivingBigBlockDataReceivedLength;
        protected int mReceivingBigBlockDataLastPacketID;
        protected byte[] mRecevingBigBlockData;

        protected bool mSendingBigBlockData = false;
        protected byte[] mBigBlockData = null;
        protected int mBigBlockByteCount = 0;
        protected int mBigBlockPacketCount = 0;

        protected int mLatchValue = 0;

        protected bool mRequestedPANForPINBlock = false;

        protected bool mPendingResponse = false;
        protected List<MTCMSRequestMessage> mRequestList = new List<MTCMSRequestMessage>();

        delegate void deviceListDispatcher(List<MTDeviceInformation> deviceList);
        delegate void updateDisplayDispatcher(string text);
        delegate void clearDispatcher();
        delegate void updateStateDispatcher(MTConnectionState state);
        delegate void userSelectionsDisptacher(string title, List<string> selectionList, long timeout);
        delegate void displayPINWindowDisptacher(long timeout);

        public MainWindow()
        {
            InitializeComponent();

            Version versionInfo = Assembly.GetExecutingAssembly().GetName().Version;

            this.Title = "MagTek CMS .NET Demo " + versionInfo.ToString();

            mDevice = new MTDevice();

            mDevice.OnDeviceList += OnDeviceList;
            mDevice.OnDeviceConnectionStateChanged += OnDeviceConnectionStateChanged;
            mDevice.OnDeviceDataString += OnDeviceDataString;
            mDevice.OnDeviceDataBytes += OnDeviceDataBytes;
            mDevice.OnDeviceResponseMessage += OnDeviceResponseMessage;
            mDevice.OnDeviceNotificationMessage += OnDeviceNotificationMessage;

            if (SubscriptionCB.Items.Count > 0)
            {
                SubscriptionCB.SelectedIndex = mSubscriptionOption;
            }

            if (MSRRequestCB.Items.Count > 0)
            {
                MSRRequestCB.SelectedIndex = mMSRRequestOption;
            }

            if (DeviceTypeCB.Items.Count > 0)
            {
                DeviceTypeCB.SelectedIndex = 0;
            }

            if (ARQCResponseCB.Items.Count > 0)
            {
                m_emvARCType = 0;
                ARQCResponseCB.SelectedIndex = 0;
            }

            if (InfoTypeCB.Items.Count > 0)
            {
                InfoTypeCB.SelectedIndex = 0;
            }

            SwipeCheckBox.IsChecked = m_enableSwipe;
            ChipCheckBox.IsChecked = m_enableChip;
            ContactlessCheckBox.IsChecked = m_enableContactless;

            updateState(MTConnectionState.Disconnected);
        }

        private void updateCardType()
        {
            bool? isSwipe = SwipeCheckBox.IsChecked;
            bool? isChip = ChipCheckBox.IsChecked;
            bool? isContactless = ContactlessCheckBox.IsChecked;

            m_enableSwipe = (isSwipe == true);
            m_enableChip = (isChip == true);
            m_enableContactless = (isContactless == true);
        }

        private byte getCardTypeValue()
        {
            byte value = 0;

            if (m_enableSwipe)
            {
                value |= 0x01;
            }

            if (m_enableChip)
            {
                value |= 0x02;
            }

            if (m_enableContactless)
            {
                value |= 0x04;
            }

            return value;
        }

        private void clearDisplay()
        {
            try
            {
                if (OutputTextBox.Dispatcher.CheckAccess())
                {
                    OutputTextBox.Clear();
                }
                else
                {
                    OutputTextBox.Dispatcher.BeginInvoke(new clearDispatcher(clearDisplay),
                                                            System.Windows.Threading.DispatcherPriority.Normal);
                }
            }
            catch (Exception)
            {
            }
        }

        private void clearMessage()
        {
            displayMessage("");
        }

        private void clearMessage2()
        {
            displayMessage2("");
        }

        private void displayMessage(string message)
        {
            try
            {
                if (MessageTextBox.Dispatcher.CheckAccess())
                {
                    MessageTextBox.Text = message;
                }
                else
                {
                    MessageTextBox.Dispatcher.BeginInvoke(new updateDisplayDispatcher(displayMessage),
                                                            System.Windows.Threading.DispatcherPriority.Normal,
                                                            new object[] { message });
                }
            }
            catch (Exception)
            {
            }
        }

        private void displayMessage2(string message)
        {
            try
            {
                if (MessageTextBox2.Dispatcher.CheckAccess())
                {
                    MessageTextBox2.Text = message;
                }
                else
                {
                    MessageTextBox2.Dispatcher.BeginInvoke(new updateDisplayDispatcher(displayMessage2),
                                                            System.Windows.Threading.DispatcherPriority.Normal,
                                                            new object[] { message });
                }
            }
            catch (Exception)
            {
            }
        }

        private void sendToDisplay(string data)
        {
            try
            {
                if (OutputTextBox.Dispatcher.CheckAccess())
                {
                    OutputTextBox.AppendText(data + "\n");
                    OutputTextBox.ScrollToEnd();
                }
                else
                {
                    OutputTextBox.Dispatcher.BeginInvoke(new updateDisplayDispatcher(sendToDisplay),
                                                            System.Windows.Threading.DispatcherPriority.Normal,
                                                            new object[] { data });
                }
            }
            catch (Exception)
            {
            }
        }

        private MTConnectionType getConnectionType()
        {
            int connectionTypeIndex = DeviceTypeCB.SelectedIndex;

            MTConnectionType connectionType = MTConnectionType.USB;

            switch (connectionTypeIndex)
            {
                case 0:
                    connectionType = MTConnectionType.USB;
                    break;
                case 1:
                    connectionType = MTConnectionType.IP;
                    break;
                case 2:
                    connectionType = MTConnectionType.Serial;
                    break;
            }

            return connectionType;
        }

        private string getAddress(MTConnectionType connectionType)
        {
            string address = "";

            switch (connectionType)
            {
                case MTConnectionType.USB:
                    MTDeviceInformation usbDevInfo = ( MTDeviceInformation ) DeviceAddressCB.SelectedValue;
                    if (usbDevInfo != null)
                    {
                        address = usbDevInfo.Address;
                    }
                    break;
                case MTConnectionType.IP:
                    address = DeviceAddressCB.Text;
                    break;
                case MTConnectionType.Serial:
                    MTDeviceInformation serialDevInfo = ( MTDeviceInformation ) DeviceAddressCB.SelectedValue;
                    if (serialDevInfo != null)
                    {
                        //address = "PORT=" + serialDevInfo.Address;
                        address = "PORT=" + serialDevInfo.Address + ",BAUDRATE=" + mBaudRate + ",DATABITS=" + mDataBits + ",PARITY=" + mParity +
                                    ",STOPBITS=" + mStopBits + ",HANDSHAKE=" + mHandshake + ",STARTINGBYTE=" + mStartingByte + ",ENDINGBYTE=" + mEndingByte +
                                    ",CRCMODE=" + ( mCRC ? "1" : "0" );
                    }
                    break;
            }

            return address;
        }

        private void scanDevices(MTConnectionType connectionType)
        {
            mDevice.requestDeviceList(connectionType);
        }

        private void connect()
        {
            if (mConnectionState == MTConnectionState.Connected)
            {
                return;
            }

            mSendingBigBlockData = false;

            mConnectionType = getConnectionType();

            mDevice.setConnectionType(mConnectionType);

            string address = getAddress(mConnectionType);

            mDevice.setAddress(address);

            mDevice.openDevice();
        }


        private void displayDeviceInformation()
        {
            MTDeviceInformation devInfo = ( MTDeviceInformation ) DeviceAddressCB.SelectedValue;

            if (devInfo != null)
            {
                string deviceInfo = "[Device Information]\n";

                deviceInfo += string.Format("Name={0}\n", devInfo.Name);
                deviceInfo += string.Format("Address={0}\n", devInfo.Address);

                if (devInfo.ProductId > 0)
                {
                    deviceInfo += string.Format("ProductID=0x{0,2:X2}\n", devInfo.ProductId);
                }
                
                sendToDisplay(deviceInfo);
            }
        }

        private void disconnect()
        {
            if (mConnectionState != MTConnectionState.Disconnected)
            {
                mDevice.closeDevice();
            }
        }

        private int sendData(string data)
        {
            int result = MTDevice.SEND_ERROR;

            if (mDevice.isDeviceConnected())
            {
                sendToDisplay("[Sending Data]");
                sendToDisplay(data);

                result = mDevice.sendDataString(data);
            }

            return result;
        }

        protected void OnDeviceList(object sender, MTConnectionType connectionType, List<MTDeviceInformation> deviceList)
        {
            updateDeviceList(deviceList);
        }

        protected void OnDeviceConnectionStateChanged(object sender, MTConnectionState state)
        {
            updateState(state);
        }

        protected void OnDeviceDataString(object sender, string dataString)
        {
            sendToDisplay("[Device Data]");
            sendToDisplay(dataString);
        }

        protected void OnDeviceDataBytes(object sender, byte[] dataBytes)
        {
        }

        protected void OnDeviceResponseMessage(object sender, MTCMSResponseMessage response)
        {
            sendToDisplay("[Device Response]");

            processResponseMessage(response);

            if (mPendingResponse)
            {
                mPendingResponse = false;
            }

            sendPendingRequests();
        }

        protected void OnDeviceNotificationMessage(object sender, MTCMSNotificationMessage notification)
        {
            //sendToDisplay("[Device Notification]");

            processNotificationMessage(notification);
        }

        private void updateState(MTConnectionState state)
        {
            mConnectionState = state;

            try
            {
                if (OutputTextBox.Dispatcher.CheckAccess())
                {
                    switch (state)
                    {
                        case MTConnectionState.Connecting:
                            updateControlStates(false);
                            OutputTextBox.Background = Brushes.Gray;
                            sendToDisplay("[Connecting....]");
                            displayDeviceInformation();
                            break;
                        case MTConnectionState.Connected:
                            resetReceivingBigBlockData();
                            updateControlStates(true);
                            OutputTextBox.Background = Brushes.White;
                            sendToDisplay("[Connected]");
                            clearMessage();
                            clearMessage2();
                            if (isAutoSubscriptionEnabled())
                            {
                                setupSubscription(true);
                            }
                            break;
                        case MTConnectionState.Disconnecting:
                            updateControlStates(false);
                            OutputTextBox.Background = Brushes.Gray;
                            sendToDisplay("[Disconnecting....]");
                            break;
                        case MTConnectionState.Disconnected:
                            updateControlStates(false);
                            OutputTextBox.Background = Brushes.Gray;
                            sendToDisplay("[Disconnected]");
                            break;
                    }
                }
                else
                {
                    OutputTextBox.Dispatcher.BeginInvoke(new updateStateDispatcher(updateState),
                                                    System.Windows.Threading.DispatcherPriority.Normal,
                                                    new object[] { state });
                }
            }
            catch (Exception ex)
            {
            }

        }

        private void setupSubscriptionDelayed(int msDelay)
        {
            Task task = Task.Factory.StartNew(async () =>
            {
                await Task.Delay(msDelay);

                setupSubscription(true);
            });
        }

        protected void updateControlStates(bool connected)
        {
            ConnectButton.IsEnabled = !connected;
            DisconnectButton.IsEnabled = connected;
            DeviceTypeCB.IsEnabled = !connected;
            DeviceAddressCB.IsEnabled = !connected;
            ScanButton.IsEnabled = !connected;

            SubscribeButton.IsEnabled = connected;
            UnsubscribeButton.IsEnabled = connected;
            MSRRequestCB.IsEnabled = connected;
            RequestMSRButton.IsEnabled = connected;
            RequestPANButton.IsEnabled = connected;
            LatchButton.IsEnabled = connected;
            UnlatchButton.IsEnabled = connected;
            InfoTypeCB.IsEnabled = connected;
            GetInfoButton.IsEnabled = connected;
            SendDataButton.IsEnabled = connected;
            EMVStartButton.IsEnabled = connected;
            EMVCancelButton.IsEnabled = connected;
            ARQCResponseCB.IsEnabled = connected;

            if (getConnectionType() == MTConnectionType.Serial)
            {
                SettingsButton.IsEnabled = !connected;
            }
            else
            {
                SettingsButton.IsEnabled = false;
            }
        }

        protected void clearDeviceList()
        {
            try
            {
                if (DeviceAddressCB.Dispatcher.CheckAccess())
                {
                    DeviceAddressCB.Items.Clear();
                }
                else
                {
                    DeviceAddressCB.Dispatcher.BeginInvoke(new clearDispatcher(clearDeviceList),
                                                            System.Windows.Threading.DispatcherPriority.Normal);
                }
            }
            catch (Exception)
            {
            }
        }

        protected void updateDeviceList(List<MTDeviceInformation> deviceList)
        {
            try
            {
                if (DeviceAddressCB.Dispatcher.CheckAccess())
                {
                    DeviceAddressCB.Items.Clear();

                    if (deviceList.Count > 0)
                    {
                        foreach (var device in deviceList)
                        {
                            DeviceAddressCB.Items.Add(device);
                        }

                        DeviceAddressCB.Visibility = Visibility.Visible;

                        DeviceAddressCB.SelectedIndex = 0;
                    }

                    DeviceAddressCB.IsEnabled = true;
                }
                else
                {
                    OutputTextBox.Dispatcher.BeginInvoke(new deviceListDispatcher(updateDeviceList),
                                                            System.Windows.Threading.DispatcherPriority.Normal,
                                                            new object[] { deviceList });
                }
            }
            catch (Exception)
            {
            }

        }

        private bool isAutoSubscriptionEnabled()
        {
            return ( mSubscriptionOption == 0 );
        }

        private bool isAutoMSRRequestEnabled()
        {
            return ( mMSRRequestOption == 0 );
        }

        private void SubscriptionCB_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            mSubscriptionOption = SubscriptionCB.SelectedIndex;
        }

        private void SubscribeButton_Click(object sender, RoutedEventArgs e)
        {
            setupSubscription(true);
        }

        private void UnsubscribeButton_Click(object sender, RoutedEventArgs e)
        {
            setupSubscription(false);
        }

        private void MSRRequestCB_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            mMSRRequestOption = MSRRequestCB.SelectedIndex;
        }

        private void RequestMSRDataButton_Click(object sender, RoutedEventArgs e)
        {
            requestMSRData();
        }

        private void RequestPANButton_Click(object sender, RoutedEventArgs e)
        {
            requestPAN();
        }

        private void DeviceTypeCB_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            DeviceAddressCB.Items.Clear();

            MTConnectionType connectionType = getConnectionType();

            switch (connectionType)
            {
                case MTConnectionType.USB:
                    DeviceAddressCB.IsEditable = false;
                    SettingsButton.IsEnabled = false;
                    scanDevices(connectionType);
                    break;
                case MTConnectionType.Serial:
                    DeviceAddressCB.IsEditable = false;
                    SettingsButton.IsEnabled = true;
                    scanDevices(connectionType);
                    break;
                case MTConnectionType.IP:
                    DeviceAddressCB.Items.Clear();
                    SettingsButton.IsEnabled = false;
                    DeviceAddressCB.IsEditable = true;
                    break;
            }
        }

        private void SettingsButton_Click(object sender, RoutedEventArgs e)
        {
            displaySerialPortSettings();
        }

        private void ScanButton_Click(object sender, RoutedEventArgs e)
        {
            clearDeviceList();

            MTConnectionType connectionType = getConnectionType();

            scanDevices(connectionType);
        }

        private void ConnectButton_Click(object sender, RoutedEventArgs e)
        {
            if (mConnectionState == MTConnectionState.Disconnected)
            {
                connect();
            }
        }

        private void DisconnectButton_Click(object sender, RoutedEventArgs e)
        {
            if (mConnectionState != MTConnectionState.Disconnected)
            {
                disconnect();
            }
        }

        private void ClearButton_Click(object sender, RoutedEventArgs e)
        {
            clearDisplay();
            clearMessage();
            clearMessage2();
        }

        private void SendDataButton_Click(object sender, RoutedEventArgs e)
        {
            string dataText = DataTextBox.Text;

            string data = dataText.Replace(" ", "");

            sendData(data);
        }

        private void EMVStartButton_Click(object sender, RoutedEventArgs e)
        {
            updateCardType();

            startEMVTransaction();
        }

        private void EMVCancelButton_Click(object sender, RoutedEventArgs e)
        {
            stopEMVTransaction();
        }

        private void ARQCResponseCB_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            m_emvARCType = ARQCResponseCB.SelectedIndex;
        }

        private void LatchButton_Click(object sender, RoutedEventArgs e)
        {
            latchControl(0);
        }

        private void UnlatchButton_Click(object sender, RoutedEventArgs e)
        {
            latchControl(1);
        }

        private void GetInfoButton_Click(object sender, RoutedEventArgs e)
        {
            int infoType = InfoTypeCB.SelectedIndex;

            switch (infoType)
            {
                case 0:
                    getMainAppInfo();
                    break;
                case 1:
                    getFWInfo();
                    break;
                case 2:
                    getBuildInfo();
                    break;
                case 3:
                    getDeviceInfo();
                    break;
                case 4:
                    getMACAddressInfo();
                    break;
                case 5:
                    getIPAddressInfo();
                    break;
                case 6:
                    getEMVL2Info();
                    break;
                case 7:
                    getDeviceStatus();
                    break;
                case 8:
                    getAllInfo();
                    break;
            }
        }

        private void displaySerialPortSettings()
        {
            try
            {
                SerialSettingsWindow dialog = new SerialSettingsWindow();

                dialog.Owner = this;

                dialog.init(mBaudRate, mDataBits, mParity, mStopBits, mHandshake, mStartingByte, mEndingByte, mCRC);

                bool? result = dialog.ShowDialog();

                if (result == true)
                {
                    mBaudRate = dialog.getBaudRate();
                    mDataBits = dialog.getDataBits();
                    mParity = dialog.getParity();
                    mStopBits = dialog.getStopBits();
                    mHandshake = dialog.getHandshake();
                    mStartingByte = dialog.getStartingByte();
                    mEndingByte = dialog.getEndingByte();
                    mCRC = dialog.getCRC();
                }
            }
            catch (Exception)
            {
            }

        }

        private void setupSubscription(bool subscribe)
        {
            sendToDisplay("[Subscribe] = " + subscribe);

            byte[] dataBytes = new byte[2];
            dataBytes[0] = subscribe ? ( byte ) 1 : ( byte ) 2;
            dataBytes[1] = 0;

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.GENERAL_COMMAND, GeneralCommandID.MESSAGE_SUBSCRIPTION, DataTypeTag.PRIMITIVE, dataBytes);

            sendToDevice(request);
        }

        private void requestMSRData()
        {
            sendToDisplay("[Send MSR Card Data Request]");

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.MSR, MSRCommandID.MSR_DATA, DataTypeTag.PRIMITIVE, null);

            sendToDevice(request);
        }

        private void requestPAN()
        {
            sendToDisplay("[Send PAN Request]");

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.PAN, PANCommandID.REQUEST_PAN, DataTypeTag.PRIMITIVE, null);

            sendToDevice(request);
        }

        private void latchControl(int value)
        {
            mLatchValue = value;

            if (mLatchValue == 0)
            {
                sendToDisplay("Latching...");
            }
            else
            {
                sendToDisplay("Unlatching...");
            }

            byte[] data = new byte[] { ( byte ) value };

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.DEVICE_CONFIGURATION, DeviceConfigCommandID.CARD_LATCH_CONTROL, DataTypeTag.PRIMITIVE, data);

            sendToDevice(request);
        }

        private void getFWInfo()
        {
            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.FIRMWARE_NUMBER, DataTypeTag.PRIMITIVE, null);

            sendToDevice(request);
        }

        private void getBuildInfo()
        {
            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.BUILD_INFO, DataTypeTag.PRIMITIVE, null);

            sendToDevice(request);
        }

        private void getDeviceInfo()
        {
            getProductName();
        }

        private void getMACAddressInfo()
        {
            byte[] data = new byte[] { 0x00 };

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.NETWORK_INFO, DataTypeTag.PRIMITIVE, data);

            sendToDevice(request);
        }

        private void getIPAddressInfo()
        {
            byte[] data = new byte[] { 0x01 };

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.NETWORK_INFO, DataTypeTag.PRIMITIVE, data);

            sendToDevice(request);
        }

        private void getMainAppInfo()
        {
            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.MAIN_SW_VERSION, DataTypeTag.PRIMITIVE, null);

            sendToDevice(request);
        }

        private void getEMVL2Info()
        {
            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.EMV_L2_VERSION, DataTypeTag.PRIMITIVE, null);

            sendToDevice(request);
        }

        private void getProductName()
        {
            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.PRODUCT_NAME, DataTypeTag.PRIMITIVE, null);

            sendToDevice(request);
        }

        private void getDeviceStatus()
        {
            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.GENERAL_COMMAND, GeneralCommandID.DEVICE_STATUS, DataTypeTag.PRIMITIVE, null);

            sendToDevice(request);
        }

        private void getAllInfo()
        {
            List<MTCMSRequestMessage> requests = new List<MTCMSRequestMessage>();

            addRequest(new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.MAIN_SW_VERSION, DataTypeTag.PRIMITIVE, null));
            addRequest(new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.FIRMWARE_NUMBER, DataTypeTag.PRIMITIVE, null));
            addRequest(new MTCMSRequestMessage(ApplicationID.DEVICE_INFORMATION, DeviceInfoCommandID.BUILD_INFO, DataTypeTag.PRIMITIVE, null));
        }

        private void addRequest(MTCMSRequestMessage request)
        {
            mRequestList.Add(request);

            sendPendingRequests();
        }

        private void sendPendingRequests()
        {
            if (mPendingResponse)
                return;

            mPendingResponse = false;

            if (mRequestList.Count > 0)
            {
                MTCMSRequestMessage request = mRequestList[0];
                mRequestList.RemoveAt(0);

                mPendingResponse = true;
                sendToDevice(request);
            }
        }

        private void startEMVTransaction()
        {
            sendToDisplay("Starting EMV Transaction...");

            clearMessage();
            clearMessage2();

            byte cardType = getCardTypeValue();

            byte[] data = new byte[19];

            System.Array.Clear(data, 0, 19);
            
            data[0] = 0x30; // General timeout
            data[1] = cardType;
            data[2] = (m_emvARCType == 2) ? (byte) 0x80 : (byte) 0x00; // Quick Chip

            byte[] amount = new byte[] { 0x00, 0x00, 0x00, 0x00, 0x15, 0x00 };
            Array.Copy(amount, 0, data, 3, 6);

            data[9] = 0x00; // Transaction Type: Purchase
            
            byte[] cashBack = new byte[] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
            Array.Copy(cashBack, 0, data, 10, 6);

            byte[] currencyCode = new byte[] { 0x08, 0x40 };
            Array.Copy(currencyCode, 0, data, 16, 2);

            data[18] = 0x02; // Notification: All status changes

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.EMV_L2, EMVL2CommandID.EMV_L2_START_TRANSACTION, DataTypeTag.PRIMITIVE, data);

            sendToDevice(request);
        }

        private void stopEMVTransaction()
        {
            sendToDisplay("Stopping EMV Transaction...");

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.EMV_L2, EMVL2CommandID.EMV_L2_CANCEL_TRANSACTION, DataTypeTag.PRIMITIVE, null);

            sendToDevice(request);
        }
 
        private void processResponseMessage(MTCMSResponseMessage response)
        {
            int applicationID = response.getApplicationID();
            int commandID = response.getCommandID();
            int resultCode = response.getResultCode();
            byte[] data = response.getData();
            
            //sendToDisplay("Response received for ApplicationID=0x" + getHexString(applicationID) + ", CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]");

            switch (applicationID)
            {
                case ApplicationID.DEVICE_INFORMATION:
                    processDeviceInfoResponse(commandID, resultCode, data);
                    break;
                case ApplicationID.DEVICE_CONFIGURATION:
                    processDeviceConfigurationResponse(commandID, resultCode, data);
                    break;
                case ApplicationID.GENERAL_COMMAND:
                    processGeneralResponse(commandID, resultCode, data);
                    break;
                case ApplicationID.MSR:
                    processMSRResponse(commandID, resultCode, data);
                    break;
                case ApplicationID.PAN:
                    processPANResponse(commandID, resultCode, data);
                    break;
                    break;
                case ApplicationID.EMV_L2:
                    processEMVL2Response(commandID, resultCode, data);
                    break;
                case ApplicationID.PIN_PAD:
                    processPINPADResponse(commandID, resultCode, data);
                    break;
            }
        }

        private void processDeviceInfoResponse(int commandID, int resultCode, byte[] data)
        {
            sendToDisplay("Device Info Application Response CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]\n");

            switch (commandID)
            {
                case DeviceInfoCommandID.PRODUCT_NAME:
                    string productName = System.Text.Encoding.UTF8.GetString(data);
                    sendToDisplay("Product Name: " + productName);
                    break;
                case DeviceInfoCommandID.FIRMWARE_NUMBER:
                    string fwInfo = System.Text.Encoding.UTF8.GetString(data);
                    sendToDisplay("Firmware Number: " + fwInfo);
                    break;
                case DeviceInfoCommandID.BUILD_INFO:
                    string buildInfo = System.Text.Encoding.UTF8.GetString(data);
                    sendToDisplay("Build Info: " + buildInfo);
                    break;
                case DeviceInfoCommandID.NETWORK_INFO:
                    if (data.Length > 1)
                    {
                        byte infoType = data[0];
                        int infoLen = data.Length - 1;
                        if (infoLen > 0)
                        {
                            byte[] infoBytes = new byte [infoLen];
                            Array.Copy(data, 1, infoBytes, 0, infoLen);
                            string infoString = System.Text.Encoding.UTF8.GetString(infoBytes);

                            switch (infoType)
                            {
                                case 0:
                                    sendToDisplay("MAC Address: " + infoString);
                                    break;
                                case 1:
                                    sendToDisplay("IP Address: " + infoString);
                                    break;
                            }
                        }
                    }
                    break;
                case DeviceInfoCommandID.MAIN_SW_VERSION:
                    string commAppInfo = System.Text.Encoding.UTF8.GetString(data);
                    sendToDisplay("Main App Version: " + commAppInfo);
                    break;
                case DeviceInfoCommandID.EMV_L2_VERSION:
                    string emvL2Info = System.Text.Encoding.UTF8.GetString(data);
                    sendToDisplay("EMV L2 Version: " + emvL2Info);
                    break;
            }
        }

        private void processGeneralResponse(int commandID, int resultCode, byte[] data)
        {
            //sendToDisplay("General Application Response CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]\n");

            switch (commandID)
            {
                case GeneralCommandID.DEVICE_STATUS:
                    processDeviceStatusResponse(data);
                    break;
                case GeneralCommandID.SEND_BIG_BLOCK_DATA:
                    processSendBigBlockDataResponse(resultCode);
                    break;
            }
        }

        private void processDeviceStatusResponse(byte[] data)
        {
//            sendToDisplay("Device Status=[" + getHexString(data) + "]\n");

            showDeviceStatus(data);
        }

        private void processSendBigBlockDataResponse(int resultCode)
        {
            if (mSendingBigBlockData)
            {
                if (resultCode == 0)
                {
                    uploadNextDataBlock();
                }
                else
                {
                    mSendingBigBlockData = false;
                }
            }
        }

        private void processMSRResponse(int commandID, int resultCode, byte[] data)
        {
            //sendToDisplay("MSR Application Response CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]\n");

            switch (commandID)
            {
                case MSRCommandID.MSR_DATA:
                    sendToDisplay("[MSR Card Data Received]");
                    processMSRData(resultCode, data);
                    break;
            }
        }

        private void processPANResponse(int commandID, int resultCode, byte [] data)
        {
            //sendToDisplay("PAN Response CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]\n");

            switch (commandID)
            {
                case PANCommandID.REQUEST_PAN:
                    processRequestPANResponse(data);
                    break;
            }
        }

        private void processDeviceConfigurationResponse(int commandID, int resultCode, byte[] data)
        {
            //sendToDisplay("Device Configuration Response CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]\n");

            if (commandID == DeviceConfigCommandID.CARD_LATCH_CONTROL)
            {
                if (mLatchValue == 0)
                {
                    sendToDisplay("[Latched]\n");
                }
                else
                {
                    sendToDisplay("[Unlatched]\n");
                }
            }
        }

        private void processEMVL2Response ( int commandID, int resultCode, byte[] data )
        {
            //sendToDisplay ( "EMV L2 Application Response CommandID=0x" + getHexString ( commandID ) + ", ResultCode=0x" + getHexString ( resultCode ) + ", Data=[" + getHexString ( data ) + "]\n" );
        }

        private void processPINPADResponse(int commandID, int resultCode, byte [] data)
        {
            //sendToDisplay("PINPAD Response CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]\n");
        }

        private void processRequestPANResponse(byte [] data)
        {
            sendToDisplay("[Response for Request PAN]");
            sendToDisplay(MTParser.getHexString(data));

            if (mRequestedPANForPINBlock)
            {
                mRequestedPANForPINBlock = false;
                getPINCVM();
            }
        }

        private void processNotificationMessage(MTCMSNotificationMessage notification)
        {
            int applicationID = notification.getApplicationID();
            int commandID = notification.getCommandID();
            int resultCode = notification.getResultCode();
            byte[] data = notification.getData();

            //sendToDisplay("Notification received for ApplicationID=0x" + getHexString(applicationID) + ", CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]");

            switch (applicationID)
            {
                case ApplicationID.GENERAL_COMMAND:
                    processGeneralNotification(commandID, resultCode, data);
                    break;
                case ApplicationID.MSR:
                    processMSRNotification(commandID, resultCode, data);
                    break;
                case ApplicationID.EMV_L2:
                    processEMVL2Notification(commandID, resultCode, data);
                    break;
                case ApplicationID.PIN_PAD:
                    processPINPADNotification(commandID, resultCode, data);
                    break;
            }
        }

        private void processGeneralNotification(int commandID, int resultCode, byte[] data)
        {
            sendToDisplay("General Application Notification CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]");

            switch (commandID)
            {
                case GeneralCommandID.DEVICE_STATUS:
                    processDeviceStatus(data);
                    break;
                case GeneralCommandID.SEND_BIG_BLOCK_DATA:
                    processBigBlockDataNotification(data);
                    break;
                case GeneralCommandID.CARD_STATUS:
                    processCardStatus(data);
                    break;
            }
        }

        private void processMSRNotification(int commandID, int resultCode, byte[] data)
        {
            //sendToDisplay("MSR Application Notification CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]");

            switch (commandID)
            {
                case MSRCommandID.MSR_CARD_DATA_AVAILABLE:
                    sendToDisplay("[MSR Card Data Available]");
                    sendToDisplay("Data =[ " + getHexString(data) + "]");
                    if (isAutoMSRRequestEnabled())
                    {
                        requestMSRData();
                    }
                    break;
                case MSRCommandID.MSR_DATA:
                    sendToDisplay("[MSR Card Data Received]");
                    processMSRData(resultCode, data);
                    break;
            }
        }

        private void processEMVL2Notification(int commandID, int resultCode, byte[] data)
        {
            //sendToDisplay ( "EMV L2 Application Notification CommandID=0x" + getHexString ( commandID ) + ", ResultCode=0x" + getHexString ( resultCode ) + ", Data=[" + getHexString ( data ) + "]\n" );

            switch ( commandID )
            {
                case EMVL2CommandID.EMV_L2_TRANSACTION_STATUS:
                    processTransactionStatus(data);
                    break;
                case EMVL2CommandID.EMV_L2_DISPLAY_MESSAGE_REQUEST:
                    processDisplayMessageRequest(data);
                    break;
                case EMVL2CommandID.EMV_L2_USER_SELECTION_REQUEST:
                    processUserSelectionRequest(data);
                    break;
                case EMVL2CommandID.EMV_L2_ARQC_MESSAGE:
                    processARQC(data);
                    break;
                case EMVL2CommandID.EMV_L2_CANCEL_TRANSACTION:
                    processCancelTransaction(data);
                    break;
                case EMVL2CommandID.EMV_L2_TRANSACTION_RESULT:
                    processTransactionResult(data);
                    break;
                case EMVL2CommandID.EMV_L2_PIN_ENTRY_SHOW_PROMPT:
                    processPINEntryShowPrompt(data);
                    break;
                case EMVL2CommandID.EMV_L2_PIN_CVM_REQUEST:
                    processPINCVMRequest();
                    break;
            }
        }

        private void processPINPADNotification(int commandID, int resultCode, byte [] data)
        {
            //sendToDisplay("PINPAD Notification CommandID=0x" + getHexString(commandID) + ", ResultCode=0x" + getHexString(resultCode) + ", Data=[" + getHexString(data) + "]\n");
        }

        private void processTransactionStatus(byte[] data)
        {
            sendToDisplay("[Transaction Status]");
            sendToDisplay(MTParser.getHexString(data));
        }

        private void processDisplayMessageRequest(byte[] data)
        {
            sendToDisplay("[Display Message Request]");

            String message = "";

            if (( data != null ) && ( data.Length > 0 ))
            {
                message = System.Text.Encoding.UTF8.GetString(data);
            }

            sendToDisplay(message);
            displayMessage(message);
        }

        private void processUserSelectionRequest(byte[] data)
        {
            if (data != null)
            {
                int dataLen = data.Length;

                if (dataLen > 2)
                {
                    byte selectionType = data[0];

                    long timeout = ((long) (data[1] & 0xFF) * 1000);

                    List<string> selectionList = getSelectionList(data, 2);

                    string title = selectionList[0];

                    selectionList.RemoveAt(0);

                    int nSelections = selectionList.Count();

                    if (nSelections > 0)
                    {
                        if (selectionType == SELECTION_TYPE_LANGUAGE)
                        {
                            for (int i = 0; i < nSelections; i++)
                            {
                                byte[] code = System.Text.Encoding.UTF8.GetBytes(selectionList[i].ToArray());
                                EMVLanguage language = EMVLanguage.GetLanguage(code);

                                if (language != null)
                                {
                                    selectionList [ i ] = language.Name;
                                }
                            }
                        }

                        timeout = 10000;

                        displayUserSelections(title, selectionList, timeout);
                    }
                }
            }
        }

        protected List<string> getSelectionList(byte [] data, int offset)
        {
            List<string> selectionList = new List<string>();

            if (data != null)
            {
                int dataLen = data.Length;

                if (dataLen >= offset)
                {
                    int start = offset;

                    for (int i = offset; i < dataLen; i++)
                    {
                        if (data [ i ] == 0x00)
                        {
                            int len = i - start;

                            if (len >= 0)
                            {
                                byte[] selectionBytes = new byte[len];

                                Array.Copy(data, start, selectionBytes, 0, len);

                                string selectionString = System.Text.Encoding.UTF8.GetString(selectionBytes);

                                selectionList.Add(selectionString);
                            }

                            start = i + 1;
                        }
                    }
                }
            }

            return selectionList;
        }

        protected void displayUserSelections(string title, List<string> selectionList, long timeout)
        {
            try
            {
                if (Dispatcher.CheckAccess())
                {
                    SelectionsWindow dialog = new SelectionsWindow();

                    dialog.Owner = this;
                    dialog.init(title, selectionList, timeout);

                    bool? result = dialog.ShowDialog();

                    int selectionIndex = (byte)dialog.getSelectedIndex();

                    dialog.Owner = null;
                    dialog = null;

                    if (result == false)
                    {
                        if (selectionIndex < 0)
                        {
                            setUserSelectionResult(SELECTION_STATUS_TIMED_OUT, ( byte ) 0);
                        }
                        else
                        {
                            setUserSelectionResult(SELECTION_STATUS_CANCELLED, ( byte ) 0);
                        }
                    }
                    else
                    {
                        setUserSelectionResult(SELECTION_STATUS_COMPLETED, ( byte ) ( selectionIndex ));
                    }
                }
                else
                {
                    Dispatcher.BeginInvoke(new userSelectionsDisptacher(displayUserSelections),
                                                            System.Windows.Threading.DispatcherPriority.Normal,
                                                            new object [] { title, selectionList, timeout });
                }
            }
            catch (Exception)
            {
            }

        }

        protected void displayPINWindow(long timeout)
        {
            try
            {
                if (Dispatcher.CheckAccess())
                {
                    PINWindow dialog = new PINWindow();

                    dialog.Owner = this;
                    dialog.init(timeout);

                    bool? result = dialog.ShowDialog();

                    string value = dialog.getValue();
                    bool timedout = dialog.isTimedout();

                    dialog.Owner = null;
                    dialog = null;

                    if (result == false)
                    {
                        if (timedout)
                        {
                            sendPINCVM(true, false, "");
                        }
                        else
                        {
                            sendPINCVM(false, true, "");
                        }
                    }
                    else
                    {
                        sendPINCVM(false, false, value);
                    }
                }
                else
                {
                    Dispatcher.BeginInvoke(new displayPINWindowDisptacher(displayPINWindow),
                                                            System.Windows.Threading.DispatcherPriority.Normal,
                                                            new object[] { timeout });

                }
            }
            catch (Exception)
            {
            }

        }

        public void setUserSelectionResult(byte status, byte selection)
        {
            sendToDisplay("[Sending Selection Result] Status=" + status + " Selection=" + selection);

            byte[] data = new byte[2];

            data [0] = status;
            data [1] = selection;

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.EMV_L2, EMVL2CommandID.EMV_L2_USER_SELECTION_RESULT, DataTypeTag.PRIMITIVE, data);

            sendToDevice(request);
        }

        private void processARQC(byte [] data)
        {
            sendToDisplay("[ARQC Received]");

            sendToDisplay(MTParser.getHexString(data));

            List<Dictionary<String, String>> parsedTLVList = MTParser.parseEMVData(data, true, "");

            if (parsedTLVList != null)
            {
                String deviceSNString = MTParser.getTagValue(parsedTLVList, "DFDF25");
                byte[] deviceSN = MTParser.getByteArrayFromHexString(deviceSNString);

                sendToDisplay("SN Bytes=" + deviceSNString);

                byte[] response = null;

                byte[] arc = null;

                if (m_emvARCType == 0)
                {
                    arc = ApprovedARC;
                }
                else if (m_emvARCType == 1)
                {
                    arc = DeclinedARC;
                }
                else if (m_emvARCType == 2)
                {
                    sendToDisplay("Quick Chip Mode (not sending Acquirer Response)");
                    return;
                }

                String macKSNString = MTParser.getTagValue(parsedTLVList, "DFDF54");
                byte[] macKSN = MTParser.getByteArrayFromHexString(macKSNString);

                String macEncryptionTypeString = MTParser.getTagValue(parsedTLVList, "DFDF55");
                byte[] macEncryptionType = MTParser.getByteArrayFromHexString(macEncryptionTypeString);

                response = buildAcquirerResponse(macKSN, macEncryptionType, deviceSN, arc, false);

                setAcquirerResponse(response);
            }
        }

        protected byte [] buildAcquirerResponse(byte [] macKSN, byte [] macEncryptionType, byte [] deviceSN, byte [] arc, bool useLengthHeader)
        {
            byte[] response = null;

            int lenMACKSN = 0;
            int lenMACEncryptionType = 0;
            int lenSN = 0;

            if (macKSN != null)
            {
                lenMACKSN = macKSN.Length;
            }

            if (macEncryptionType != null)
            {
                lenMACEncryptionType = macEncryptionType.Length;
            }

            if (deviceSN != null)
            {
                lenSN = deviceSN.Length;
            }

            byte[] macKSNTag = new byte[] { (byte)0xDF, (byte)0xDF, 0x54, (byte)lenMACKSN };
            byte[] macEncryptionTypeTag = new byte[] { (byte)0xDF, (byte)0xDF, 0x55, (byte)lenMACEncryptionType };
            byte[] snTag = new byte[] { (byte)0xDF, (byte)0xDF, 0x25, (byte)lenSN };
            byte[] container = new byte[] { (byte)0xFA, 0x06, 0x70, 0x04 };

            int lenTLV = 2 + macKSNTag.Length + lenMACKSN + macEncryptionTypeTag.Length + lenMACEncryptionType + snTag.Length + lenSN + container.Length + arc.Length;

            if (useLengthHeader)
                lenTLV += 2;

            int lenPadding = 0;

            if (( lenTLV % 8 ) > 0)
            {
                lenPadding = ( 8 - lenTLV % 8 );
            }

            int lenData = lenTLV + lenPadding + 4;

            response = new byte [ lenData ];

            int i = 0;

            if (useLengthHeader)
            {
                response [ i++ ] = ( byte ) ( ( ( lenData - 2 ) >> 8 ) & 0xFF );
                response [ i++ ] = ( byte ) ( ( lenData - 2 ) & 0xFF );
            }

            response [ i++ ] = ( byte ) 0xF9;
            response [ i++ ] = ( byte ) ( lenTLV - 2 );
            Array.Copy(macKSNTag, 0, response, i, macKSNTag.Length);
            i += macKSNTag.Length;
            Array.Copy(macKSN, 0, response, i, macKSN.Length);
            i += macKSN.Length;
            Array.Copy(macEncryptionTypeTag, 0, response, i, macEncryptionTypeTag.Length);
            i += macEncryptionTypeTag.Length;
            Array.Copy(macEncryptionType, 0, response, i, macEncryptionType.Length);
            i += macEncryptionType.Length;
            Array.Copy(snTag, 0, response, i, snTag.Length);
            i += snTag.Length;
            Array.Copy(deviceSN, 0, response, i, deviceSN.Length);
            i += deviceSN.Length;
            Array.Copy(container, 0, response, i, container.Length);
            i += container.Length;

            if (arc != null)
            {
                Array.Copy(arc, 0, response, i, arc.Length);
            }

            return response;
        }

        private void setAcquirerResponse(byte [] response)
        {
            sendToDisplay("[Sending Acquirer Response]\n" + MTParser.getHexString(response));

            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.EMV_L2, EMVL2CommandID.EMV_L2_ACQUIRER_RESPONSE, DataTypeTag.PRIMITIVE, response);

            if (response.Length > getBigBlockDataSize())
            {
                sendBigBlocksToDevice(request);
            }
            else
            {
                sendToDevice(request);
            }

        }

        private void processCancelTransaction(byte [] data)
        {
            sendToDisplay("[Cancel Transaction Status]");
            sendToDisplay(MTParser.getHexString(data));
        }

        private void processTransactionResult(byte [] data)
        {
            sendToDisplay("[Transaction Result]");

            sendToDisplay(MTParser.getHexString(data));

            if (data != null)
            {
                if (data.Length > 0)
                {
                    bool signatureRequired = (data[0] != 0);

                    int lenBatchData = data.Length - 3;
                    if (lenBatchData > 0)
                    {
                        byte[] batchData = new byte[lenBatchData];

                        Array.Copy(data, 3, batchData, 0, lenBatchData);

                        sendToDisplay("(Parsed Batch Data)");

                        List<Dictionary<String, String>> parsedTLVList = MTParser.parseTLVData(batchData);

                        displayParsedTLV(parsedTLVList);

                        bool approved = false;

                        String statusString = MTParser.getTagValue(parsedTLVList, "dfdf1a");
                        byte[] statusValue = MTParser.getByteArrayFromHexString(statusString);


                        if (statusValue != null)
                        {
                            if (statusValue.Length > 0)
                            {
                                if (statusValue [ 0 ] == 0)
                                {
                                    approved = true;
                                }
                            }
                        }

                        if (approved)
                        {
                            if (signatureRequired)
                            {
                                displayMessage2("( Signature Required )");
                            }
                            else
                            {
                                displayMessage2("( No Signature Required )");
                            }
                        }
                    }
                }
            }
        }

        private void processPINEntryShowPrompt(byte [] data)
        {
            sendToDisplay("[PIN Entry Show Prompt]");

            sendToDisplay(MTParser.getHexString(data));

            if (data != null)
            {
                if (data.Length > 0)
                {
                    byte promptID = data[0];

                    switch (promptID)
                    {
                        case 0x09:
                            //displayMessage("[ENTER PIN]");
                            break;
                    }

                    //mPINEntryDisplay = "";
                    clearMessage2();
                }
            }
        }

        private void processPINCVMRequest()
        {
            sendToDisplay("[PIN CVM Request]");

            mRequestedPANForPINBlock
                = true;
            requestPAN();
        }

        private void getPINCVM()
        {
            sendToDisplay("[Get PIN CVM]");

            displayPINWindow(30000);
        }

        private void sendPINCVM(bool timedout, bool cancelled, string value)
        {
            if (timedout)
            {
                sendToDisplay("(Operation timed out)");
                sendPINCVMResponse(0x03, null); // Operation timed out
            }
            else if (cancelled)
            {
                sendToDisplay("(User cancelled operation)");
                sendPINCVMResponse(0x02, null); // User cancelled operation
            }
            else
            {
                int len = 0;

                if (value != null)
                {
                    len = value.Length;
                }

                if (len > 0)
                {
                    byte[] valueBytes = System.Text.Encoding.UTF8.GetBytes(value);


                    sendPINCVMResponse(0x05, valueBytes); 
                }
                else
                {
                    sendToDisplay("(PIN Bypass selected)");
                    sendPINCVMResponse(0x05, null); // PIN Bypass selected
                }
            }
        }

        private void sendPINCVMResponse(byte result, byte[] pinData)
        {
            sendToDisplay("[Send PIN CVM Response]");

            int len = 4;

            if (pinData != null)
            {
                len += pinData.Length;
            }

            byte [] data = new byte[len];

            data[0] = 0xDF;
            data[1] = 0x45;
            data[2] = 0x01;
            data[3] = result;

            if (len > 4)
            {
                Array.Copy(pinData, 0, data, 4, len - 4);
            }

            MTCMSResponseMessage response = new MTCMSResponseMessage(ApplicationID.EMV_L2, EMVL2CommandID.EMV_L2_PIN_CVM_RESPONSE, 0, DataTypeTag.CONSTRUCTIVE, data);

            sendToDevice(response);
        }

        private void processDeviceStatus(byte[] data)
        {
            sendToDisplay("Device Status=[" + getHexString(data) + "]\n");
        }

        private void processCardStatus(byte[] data)
        {
            sendToDisplay("Card Status=[" + getHexString(data) + "]\n");

            if (data != null)
            {
                int len = data.Length;

                if (len >= 2)
                {
                    byte present = data[0];
                    byte type = data[1];

                    sendToDisplay("Card Present=[" + present + "]\n");
                    sendToDisplay("Card Type=[" + type + "]\n");
                }
            }
        }

        private void resetReceivingBigBlockData()
        {
            mReceivingBigBlockDataTotalLength = 0;
            mReceivingBigBlockDataReceivedLength = 0;
            mReceivingBigBlockDataLastPacketID = 0;
            mRecevingBigBlockData = null;
        }

        private void processBigBlockDataNotification(byte[] data)
        {
            //sendToDisplay("Big Block Data=[" + getHexString(data) + "]\n");

            if (data != null)
            {
                int len = data.Length;

                if (len >= 2)
                {
                    int packetID = (int) (data[1] << 8) + (int) data[0];

                    if (packetID == 0)
                    {
                        mReceivingBigBlockDataLastPacketID = 0;
                        mReceivingBigBlockDataReceivedLength = 0;

                        if (len >= 4)
                        {
                            int lenData = (int)(data[3] << 8) + (int)data[2];

                            if (lenData > 0)
                            {
                                if (len >= (lenData + 4))
                                {
                                    byte[] dataBytes = new byte[lenData];
                                    Array.Copy(data, 4, dataBytes, 0, lenData);

                                    int totalLen = 0;

                                    for (int i = 0; i < lenData; i++)
                                    { 
                                        totalLen += (((int) (dataBytes[i] & 0xFF)) << (i * 8));
                                    }

                                    if (totalLen > 0)
                                    {
                                        mRecevingBigBlockData = new byte[totalLen];
                                        Array.Clear(mRecevingBigBlockData, 0, totalLen);

                                        mReceivingBigBlockDataTotalLength = totalLen;
                                    }
                                }
                                else
                                {
                                    // Insufficient data available
                                }
                            }
                        }
                    }
                    else 
                    {
                        if (packetID == (mReceivingBigBlockDataLastPacketID + 1))
                        {
                            mReceivingBigBlockDataLastPacketID = packetID;

                            if (len >= 4)
                            {
                                int lenData = (int)(data[3] << 8) + (int)data[2];

                                if (lenData > 0)
                                {
                                    if (len >= (lenData + 4))
                                    {
                                        Array.Copy(data, 4, mRecevingBigBlockData, mReceivingBigBlockDataReceivedLength, lenData);

                                        mReceivingBigBlockDataReceivedLength += lenData;

                                        if (mReceivingBigBlockDataReceivedLength >= mReceivingBigBlockDataTotalLength)
                                        {
                                            MTCMSMessage message = new MTCMSMessage(mRecevingBigBlockData);

                                            if (message.getMessageType() == 2)
                                            {
                                                MTCMSResponseMessage response = new MTCMSResponseMessage(message.getApplicationID(), message.getCommandID(), message.getResultCode(), message.getDataTag(), message.getData());

                                                OnDeviceResponseMessage(this, response);
                                            }
                                            else if (message.getMessageType() == 3)
                                            {
                                                MTCMSNotificationMessage notification = new MTCMSNotificationMessage(message.getApplicationID(), message.getCommandID(), message.getResultCode(), message.getDataTag(), message.getData());

                                                OnDeviceNotificationMessage(this, notification);
                                            }
                                        }
                                    }
                                    else
                                    {
                                        // Insufficient data available
                                    }
                                }
                            }
                        }
                        else
                        { 
                            // Unexpected packetID received
                        }
                    }
                }
            }
        }

        private void processMSRData(int resultCode, byte[] data)
        {
            sendToDisplay("MSR Data=[" + getHexString(data) + "]\n");

            showMSRData(data);
        }

        private void displayParsedTLV(List<Dictionary<string, string>> parsedTLVList)
        {
            if (parsedTLVList != null)
            {
                foreach (Dictionary<String, String> map in parsedTLVList)
                {
                    string tagString;
                    string valueString;

                    if (map.TryGetValue("tag", out tagString))
                    {
                        if (map.TryGetValue("value", out valueString))
                        {
                            sendToDisplay("  " + tagString + "=" + valueString);
                        }
                    }
                }
            }
        }   

        private void showDeviceStatus(byte[] data)  
        {
            List<Dictionary<String, String>> parsedTLVList = MTParser.parseTLVData(data);

            if (parsedTLVList != null)
            {
                string deviceStatus = "";

                deviceStatus += string.Format("Device.State={0}\n", MTParser.getTagValue(parsedTLVList, "DF50"));
                deviceStatus += string.Format("Device.Status={0}\n", MTParser.getTagValue(parsedTLVList, "DF51"));
                deviceStatus += string.Format("Device.Centificate+Key.Status={0}\n", MTParser.getTagValue(parsedTLVList, "DF52"));

                sendToDisplay(deviceStatus);
            }
        }

        private void showMSRData(byte[] data)
        {
            sendToDisplay("MSR Data:\n");

            List<Dictionary<String, String>> parsedTLVList = MTParser.parseTLVData(data);

            if (parsedTLVList != null)
            {
                string msrData = "";

                msrData += string.Format("Track1.Masked={0}\n", MTParser.getTagTextValue(parsedTLVList, "DFDF31"));
                msrData += string.Format("Track2.Masked={0}\n", MTParser.getTagTextValue(parsedTLVList, "DFDF33"));
                msrData += string.Format("Track3.Masked={0}\n", MTParser.getTagTextValue(parsedTLVList, "DFDF35"));

                msrData += string.Format("Track1.Status={0}\n", MTParser.getTagValue(parsedTLVList, "DFDF36"));
                msrData += string.Format("Track2.Status={0}\n", MTParser.getTagValue(parsedTLVList, "DFDF38"));
                msrData += string.Format("Track3.Status={0}\n", MTParser.getTagValue(parsedTLVList, "DFDF3A"));

                msrData += string.Format("MagnePrint.Status={0}\n", MTParser.getTagValue(parsedTLVList, "DFDF43"));

                msrData += string.Format("KSN={0}\n", MTParser.getTagValue(parsedTLVList, "DFDF56"));
                msrData += string.Format("Card.EncodeType={0}\n", MTParser.getTagValue(parsedTLVList, "DFDF4F"));
                msrData += string.Format("EncryptionType={0}\n", MTParser.getTagValue(parsedTLVList, "DFDF51"));

                String encryptedDataHexString = MTParser.getTagValue(parsedTLVList, "DFDF59");
                msrData += string.Format("EncryptedData={0}\n", encryptedDataHexString);

/*
                byte[] encryptedDataBytes = MTParser.getByteArrayFromHexString(encryptedDataHexString);

                List<Dictionary<String, String>> encryptedDataTLVList = MTParser.parseTLVData(encryptedDataBytes);

                if (encryptedDataTLVList != null)
                {
                    String encryptedTrackHexString = MTParser.getTagValue(encryptedDataTLVList, "FA");

                    byte[] encryptedTrackBytes = MTParser.getByteArrayFromHexString(encryptedTrackHexString);

                    List<Dictionary<String, String>> encryptedTrackTLVList = MTParser.parseTLVData(encryptedTrackBytes);

                    if (encryptedTrackTLVList != null)
                    {
                        msrData += string.Format("Track1.Encrypted={0}\n", MTParser.getTagValue(encryptedTrackTLVList, "DF41"));
                        msrData += string.Format("Track2.Encrypted={0}\n", MTParser.getTagValue(encryptedTrackTLVList, "DF42"));
                        msrData += string.Format("Track3.Encrypted={0}\n", MTParser.getTagValue(encryptedTrackTLVList, "DF43"));
                        msrData += string.Format("MagnePrint={0}\n", MTParser.getTagValue(encryptedTrackTLVList, "DF44"));
                    }
                }
*/

                sendToDisplay(msrData);
            }
        }

        private int sendToDevice(MTCMSMessage message)
        {
            byte[] data = message.getMessageBytes();

            string dataHexString = getHexString(data);

            sendToDisplay("Sending to device: [" + dataHexString + "]");

            int result = mDevice.sendMTCMSMessage(message);

            sendToDisplay("Send result=" + result);

            return result;
        }

        private int sendBigBlocksToDevice(MTCMSMessage message)
        {
            if (mSendingBigBlockData)
            {
                return -1;
            }

            mBigBlockData = message.getMessageBytes();

            mBigBlockByteCount = 0;
            mBigBlockPacketCount = 0;

            sendBigBlockDataLength();

            return 0;
        }

        private int getBigBlockDataLength()
        {
            int dataLen = 0;

            if (mBigBlockData != null)
            {
                dataLen = mBigBlockData.Length; ;
            }

            return dataLen;
        }

        private void sendBigBlockDataLength()
        {
            int totalDataLen = getBigBlockDataLength();

            if (totalDataLen > 0)
            {
                mSendingBigBlockData = true;

                byte[] packetIDBytes = getPacketIDByteArray(0);
                byte[] valueLenBytes = getLengthByteArray(2, 4);
                byte[] valueBytes = getLengthByteArray(4, totalDataLen);

                byte[] dataBytes = getArrayBytes(packetIDBytes, valueLenBytes, valueBytes);

                mBigBlockPacketCount = 1;

                sendBigBlockDataCommand(dataBytes);
            }
        }

        private void sendBigBlockDataCommand(byte [] data)
        {
            MTCMSRequestMessage request = new MTCMSRequestMessage(ApplicationID.GENERAL_COMMAND, GeneralCommandID.SEND_BIG_BLOCK_DATA, DataTypeTag.PRIMITIVE, data);

            sendToDevice(request);
        }

        private byte [] buildBigBlockDataBytes(int packetID, byte [] value)
        {
            byte[] dataBytes = null;

            if (value != null)
            {
                byte[] packetIDBytes = getPacketIDByteArray(packetID);
                byte[] valueLenBytes = getLengthByteArray(2, value.Length);

                dataBytes = getArrayBytes(packetIDBytes, valueLenBytes, value);
            }

            return dataBytes;
        }

        private int getBigBlockDataSize()
        {
            int blockSize = BIG_BLOCK_DATA_SIZE;

            if (mConnectionType == MTConnectionType.USB)
            {
                blockSize = BIG_BLOCK_DATA_SMALL_SIZE;
            }

            return blockSize;
        }

        private bool uploadNextDataBlock()
        {
            int totalDataLen = getBigBlockDataLength();

            if (mBigBlockByteCount < totalDataLen)
            {
                int dataLen = totalDataLen - mBigBlockByteCount;

                int blockSize = getBigBlockDataSize();

                if (dataLen > blockSize)
                {
                    dataLen = blockSize;
                }

                byte[] data = new byte[dataLen];

                if (data != null)
                {
                    Array.Copy(mBigBlockData, mBigBlockByteCount, data, 0, dataLen);

                    byte[] dataBytes = buildBigBlockDataBytes(mBigBlockPacketCount, data);

                    mBigBlockPacketCount++;

                    mBigBlockByteCount += dataLen;

                    sendBigBlockDataCommand(dataBytes);
                }
            }
            else
            {
                mSendingBigBlockData = false;

                return false;
            }

            return true;
        }

        private byte [] getLengthByteArray(int lenBytesNeeded, int len)
        {
            byte[] lengthArray;

            lengthArray = new byte [ lenBytesNeeded ];
            Array.Clear(lengthArray, 0, lengthArray.Length);
            if (len > 0)
            {
                for (int i = 0; i < lenBytesNeeded; i++)
                {
                    lengthArray [ i ] = ( byte ) ( ( len >> ( i * 8 ) ) & ( 0xFF ) );
                }
            }

            return lengthArray;
        }

        private byte [] getPacketIDByteArray(int id)
        {
            return getLengthByteArray(2, id);
        }

        private byte [] getArrayBytes(byte [] array1, byte [] array2, byte [] array3)
        {
            int len = 0;

            if (array1 != null)
            {
                len += array1.Length;
            }

            if (array2 != null)
            {
                len += array2.Length;
            }

            if (array3 != null)
            {
                len += array3.Length;
            }

            byte[] valueBytes = null;

            if (len > 0)
            {
                int i = 0;

                valueBytes = new byte [ len ];

                if (array1 != null)
                {
                    Array.Copy(array1, 0, valueBytes, i, array1.Length);
                    i += array1.Length;
                }

                if (array2 != null)
                {
                    Array.Copy(array2, 0, valueBytes, i, array2.Length);
                    i += array2.Length;
                }

                if (array3 != null)
                {
                    Array.Copy(array3, 0, valueBytes, i, array3.Length);
                    i += array3.Length;
                }
            }

            return valueBytes;
        }


        public static string getHexString(int value)
        {
            byte byteValue = (byte)(value & 0xFF);

            return byteValue.ToString("X2");
        }

        public static string getHexString(byte[] data)
        {
            return MTParser.getHexString(data);
        }

        public static byte[] getByteArrayFromHexString(string str)
        {
            return MTParser.getByteArrayFromHexString(str);
        }
    }
}
