﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;

namespace MTCMSNETDemo
{
    /// <summary>
    /// Interaction logic for SerialSettingsWindow.xaml
    /// </summary>
    public partial class SerialSettingsWindow : Window
    {
        protected string mBaudRate;
        protected string mDataBits;
        protected string mParity;
        protected string mStopBits;
        protected string mHandshake;
        protected string mStartingByte;
        protected string mEndingByte;
        protected bool mCRC;

        protected bool mLoading = false;

        public SerialSettingsWindow()
        {
            mLoading = true;

            InitializeComponent();
            
            mLoading = false;
        }

        public void init(string baudRate, string dataBits, string parity, string stopBits, string handshake, string startingByte, string endingByte, bool crc)
        {
            mBaudRate = baudRate;
            mDataBits = dataBits;
            mParity = parity;
            mStopBits = stopBits;
            mHandshake = handshake;
            mStartingByte = startingByte;
            mEndingByte = endingByte;
            mCRC = crc;

            updateDisplay();
        }

        public string getBaudRate()
        {
            return mBaudRate;
        }

        public string getDataBits()
        {
            return mDataBits;
        }

        public string getParity()
        {
            return mParity;
        }

        public string getStopBits()
        {
            return mStopBits;
        }

        public string getHandshake()
        {
            return mHandshake;
        }

        public string getStartingByte()
        {
            return mStartingByte;
        }

        public string getEndingByte()
        {
            return mEndingByte;
        }

        public bool getCRC()
        {
            return mCRC;
        }

        protected void updateDisplay()
        {
            mLoading = true;

            int index = findItemIndex(BaudRateCB.Items, mBaudRate);
            if (index >= 0)
            {
                BaudRateCB.SelectedIndex = index;
            }

            index = findItemIndex(DataBitsCB.Items, mDataBits);
            if (index >= 0)
            {
                DataBitsCB.SelectedIndex = index;
            }

            if (mParity.StartsWith("NONE"))
            {
                ParityCB.SelectedIndex = 0;
            }
            else if (mParity.StartsWith("EVEN"))
            {
                ParityCB.SelectedIndex = 1;
            }
            else if (mParity.StartsWith("ODD"))
            {
                ParityCB.SelectedIndex = 2;
            }
            else if (mParity.StartsWith("SPACE"))
            {
                ParityCB.SelectedIndex = 3;
            }
            else if (mParity.StartsWith("MARK"))
            {
                ParityCB.SelectedIndex = 4;
            }

            index = findItemIndex(StopBitsCB.Items, mStopBits);
            if (index >= 0)
            {
                StopBitsCB.SelectedIndex = index;
            }

            if (mHandshake.StartsWith("NONE"))
            {
                FlowControlCB.SelectedIndex = 0;
            }
            else if (mHandshake.StartsWith("RTS"))
            {
                FlowControlCB.SelectedIndex = 1;
            }
            else if (mHandshake.StartsWith("XONXOFF"))
            {
                FlowControlCB.SelectedIndex = 2;
            }

            if (string.IsNullOrWhiteSpace(mStartingByte))
            {
                StartingByteCB.SelectedIndex = 0;
            }
            else if (mStartingByte.StartsWith("0x02"))
            {
                StartingByteCB.SelectedIndex = 1;
            }
            else
            {
                StartingByteCB.SelectedIndex = 2;
                StartingByteTextBox.Text = mStartingByte;
            }

            if (string.IsNullOrWhiteSpace(mEndingByte))
            {
                EndingByteCB.SelectedIndex = 0;
            }
            else if (mEndingByte.StartsWith("0x03"))
            {
                EndingByteCB.SelectedIndex = 1;
            }
            else if (mEndingByte.StartsWith("0x0A"))
            {
                EndingByteCB.SelectedIndex = 2;
            }
            else if (mEndingByte.StartsWith("0x0D"))
            {
                EndingByteCB.SelectedIndex = 3;
            }
            else
            {
                EndingByteCB.SelectedIndex = 4;
                EndingByteTextBox.Text = mEndingByte;
            }

            EnableCRCCheckBox.IsChecked = mCRC;

            mLoading = false;
        }

        protected void getValues()
        {
            int index = BaudRateCB.SelectedIndex;
            ListBoxItem item = (ListBoxItem) BaudRateCB.Items.GetItemAt(index);
            mBaudRate = (string) item.Content;

            index = DataBitsCB.SelectedIndex;
            item = (ListBoxItem) DataBitsCB.Items.GetItemAt(index);
            mDataBits = (string)item.Content;

            index = ParityCB.SelectedIndex;
            switch (index)
            { 
                case 0:
                    mParity = "NONE";
                    break;
                case 1:
                    mParity = "EVEN";
                    break;
                case 2:
                    mParity = "ODD";
                    break;
                case 3:
                    mParity = "SPACE";
                    break;
                case 4:
                    mParity = "MARK";
                    break;
            }

            index = StopBitsCB.SelectedIndex;
            item = (ListBoxItem)StopBitsCB.Items.GetItemAt(index);
            mStopBits = (string)item.Content;

            index = FlowControlCB.SelectedIndex;
            switch (index)
            {
                case 0:
                    mHandshake = "NONE";
                    break;
                case 1:
                    mHandshake = "RTS";
                    break;
                case 2:
                    mHandshake = "XONXOFF";
                    break;
            }

            index = StartingByteCB.SelectedIndex;
            switch (index)
            {
                case 0:
                    mStartingByte = "";
                    break;
                case 1:
                    mStartingByte = "0x02";
                    break;
                case 2:
                    mStartingByte = StartingByteTextBox.Text;
                    break;
            }

            index = EndingByteCB.SelectedIndex;
            switch (index)
            {
                case 0:
                    mEndingByte = "";
                    break;
                case 1:
                    mEndingByte = "0x03";
                    break;
                case 2:
                    mEndingByte = "0x0A";
                    break;
                case 3:
                    mEndingByte = "0x0D";
                    break;
                case 4:
                    mEndingByte = EndingByteTextBox.Text;
                    break;
            }

            mCRC = (bool)EnableCRCCheckBox.IsChecked;
        }

        protected int findItemIndex(ItemCollection items, string target)
        {
            for (int i = 0; i < items.Count; i++)
            { 
                if (items.GetItemAt(i).ToString().Contains(target))
                {
                    return i;
                }
            }

            return -1;
        }

        private void StartingByteCB_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (mLoading)
                return;

            if (StartingByteCB.SelectedIndex != 2)
            {
                StartingByteTextBox.Text = "";
            }
        }

        private void EndingByteCB_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (mLoading)
                return;

            if (EndingByteCB.SelectedIndex != 4)
            {
                EndingByteTextBox.Text = "";
            }
        }

        private void OKButton_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                getValues();

                this.DialogResult = true;

                this.Close();
            }
            catch (Exception)
            {
            }
        }

        private void CancelButton_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                this.DialogResult = false;

                this.Close();
            }
            catch (Exception)
            {
            }
        }
    }
}
